@@ -1,3 +1,4 @@
+from __future__ import absolute_import, print_function, division
 import unittest
 
 import numpy
@@ -14,30 +15,12 @@ from theano.tests.test_rop import break_op
 from theano.tests import unittest_tools as utt
 from theano import config
 
-from theano.tensor.nlinalg import ( MatrixInverse,
-                                    matrix_inverse,
-                                    MatrixPinv,
-                                    pinv,
-                                    AllocDiag,
-                                    alloc_diag,
-                                    ExtractDiag,
-                                    extract_diag,
-                                    diag,
-                                    trace,
-                                    Det,
-                                    det,
-                                    Eig,
-                                    eig,
-                                    Eigh,
-                                    EighGrad,
-                                    eigh,
-                                    matrix_dot,
-                                    _zero_disconnected,
-                                    qr,
-                                    matrix_power,
-                                    norm,
-                                    svd
-                                    )
+from theano.tensor.nlinalg import (
+    MatrixInverse, matrix_inverse, MatrixPinv, pinv,
+    AllocDiag, alloc_diag, ExtractDiag, extract_diag, diag,
+    trace, Det, det, Eig, eig, Eigh, EighGrad, eigh,
+    matrix_dot, _zero_disconnected, qr, matrix_power,
+    norm, svd, TensorInv, tensorinv, tensorsolve)
 from nose.plugins.attrib import attr
 
 from nose.plugins.skip import SkipTest
@@ -126,7 +109,7 @@ def test_qr_modes():
     n_qr = numpy.linalg.qr(a)
     assert _allclose(n_qr, t_qr)
 
-    for mode in ["reduced", "r", "raw", "full", "economic"]:
+    for mode in ["reduced", "r", "raw"]:
         f = function([A], qr(A, mode))
         t_qr = f(a)
         n_qr = numpy.linalg.qr(a, mode)
@@ -159,6 +142,52 @@ def test_svd():
     assert _allclose(n_t, t_t)
 
 
+def test_tensorsolve():
+    rng = numpy.random.RandomState(utt.fetch_seed())
+
+    A = tensor.tensor4("A", dtype=theano.config.floatX)
+    B = tensor.matrix("B", dtype=theano.config.floatX)
+    X = tensorsolve(A, B)
+    fn = function([A, B], [X])
+
+    # slightly modified example from numpy.linalg.tensorsolve docstring
+    a = numpy.eye(2 * 3 * 4).astype(theano.config.floatX)
+    a.shape = (2 * 3, 4, 2, 3 * 4)
+    b = rng.rand(2 * 3, 4).astype(theano.config.floatX)
+
+    n_x = numpy.linalg.tensorsolve(a, b)
+    t_x = fn(a, b)
+    assert _allclose(n_x, t_x)
+
+    # check the type upcast now
+    C = tensor.tensor4("C", dtype='float32')
+    D = tensor.matrix("D", dtype='float64')
+    Y = tensorsolve(C, D)
+    fn = function([C, D], [Y])
+
+    c = numpy.eye(2 * 3 * 4, dtype='float32')
+    c.shape = (2 * 3, 4, 2, 3 * 4)
+    d = rng.rand(2 * 3, 4).astype('float64')
+    n_y = numpy.linalg.tensorsolve(c, d)
+    t_y = fn(c, d)
+    assert _allclose(n_y, t_y)
+    assert n_y.dtype == Y.dtype
+
+    # check the type upcast now
+    E = tensor.tensor4("E", dtype='int32')
+    F = tensor.matrix("F", dtype='float64')
+    Z = tensorsolve(E, F)
+    fn = function([E, F], [Z])
+
+    e = numpy.eye(2 * 3 * 4, dtype='int32')
+    e.shape = (2 * 3, 4, 2, 3 * 4)
+    f = rng.rand(2 * 3, 4).astype('float64')
+    n_z = numpy.linalg.tensorsolve(e, f)
+    t_z = fn(e, f)
+    assert _allclose(n_z, t_z)
+    assert n_z.dtype == Z.dtype
+
+
 def test_inverse_singular():
     singular = numpy.array([[1, 0, 0]] + [[0, 1, 0]] * 2,
                            dtype=theano.config.floatX)
@@ -374,9 +403,9 @@ class test_Eig(utt.InferShapeTester):
         super(test_Eig, self).setUp()
         self.rng = numpy.random.RandomState(utt.fetch_seed())
         self.A = theano.tensor.matrix(dtype=self.dtype)
-        X = numpy.asarray(self.rng.rand(5, 5),
-                          dtype=self.dtype)
-        self.S = X.dot(X.T)
+        self.X = numpy.asarray(self.rng.rand(5, 5),
+                               dtype=self.dtype)
+        self.S = self.X.dot(self.X.T)
 
     def test_infer_shape(self):
         A = self.A
@@ -410,21 +439,21 @@ class test_Eigh(test_Eig):
                                   vl * numpy.sign(vl[0, :]))
 
     def test_grad(self):
-        S = self.S
-        utt.verify_grad(lambda x: self.op(x)[0], [S], rng=self.rng)
-        utt.verify_grad(lambda x: self.op(x)[1], [S], rng=self.rng)
-        utt.verify_grad(lambda x: self.op(x, 'U')[0], [S], rng=self.rng)
-        utt.verify_grad(lambda x: self.op(x, 'U')[1], [S], rng=self.rng)
+        X = self.X
+        # We need to do the dot inside the graph because Eigh needs a
+        # matrix that is hermitian
+        utt.verify_grad(lambda x: self.op(x.dot(x.T))[0], [X], rng=self.rng)
+        utt.verify_grad(lambda x: self.op(x.dot(x.T))[1], [X], rng=self.rng)
+        utt.verify_grad(lambda x: self.op(x.dot(x.T), 'U')[0], [X], rng=self.rng)
+        utt.verify_grad(lambda x: self.op(x.dot(x.T), 'U')[1], [X], rng=self.rng)
 
 
 class test_Eigh_float32(test_Eigh):
     dtype = 'float32'
 
-    @utt.AttemptManyTimes(n_attempts=3, n_req_successes=2)
     def test_uplo(self):
         super(test_Eigh_float32, self).test_uplo()
 
-    @utt.AttemptManyTimes(n_attempts=3, n_req_successes=2)
     def test_grad(self):
         super(test_Eigh_float32, self).test_grad()
 
@@ -516,3 +545,41 @@ class T_NormTests(unittest.TestCase):
             t_n = f(A[2][i])
             n_n = numpy.linalg.norm(A[2][i], A[3][i])
             assert _allclose(n_n, t_n)
+
+
+class test_TensorInv(utt.InferShapeTester):
+    def setUp(self):
+        super(test_TensorInv, self).setUp()
+        self.A = tensor.tensor4("A", dtype=theano.config.floatX)
+        self.B = tensor.tensor3("B", dtype=theano.config.floatX)
+        self.a = numpy.random.rand(4, 6, 8, 3).astype(theano.config.floatX)
+        self.b = numpy.random.rand(2, 15, 30).astype(theano.config.floatX)
+        self.b1 = numpy.random.rand(30, 2, 15).astype(theano.config.floatX)  # for ind=1 since we need prod(b1.shape[:ind]) == prod(b1.shape[ind:])
+
+    def test_infer_shape(self):
+        A = self.A
+        Ai = tensorinv(A)
+        self._compile_and_check([A],  # theano.function inputs
+                                [Ai],  # theano.function outputs
+                                [self.a],  # value to substitute
+                                TensorInv)
+
+    def test_eval(self):
+        A = self.A
+        Ai = tensorinv(A)
+        n_ainv = numpy.linalg.tensorinv(self.a)
+        tf_a = function([A], [Ai])
+        t_ainv = tf_a(self.a)
+        assert _allclose(n_ainv, t_ainv)
+
+        B = self.B
+        Bi = tensorinv(B)
+        Bi1 = tensorinv(B, ind=1)
+        n_binv = numpy.linalg.tensorinv(self.b)
+        n_binv1 = numpy.linalg.tensorinv(self.b1, ind=1)
+        tf_b = function([B], [Bi])
+        tf_b1 = function([B], [Bi1])
+        t_binv = tf_b(self.b)
+        t_binv1 = tf_b1(self.b1)
+        assert _allclose(t_binv, n_binv)
+        assert _allclose(t_binv1, n_binv1)
